//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of NVIDIA CORPORATION nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Copyright (c) 2008-2021 NVIDIA Corporation. All rights reserved.
// Copyright (c) 2004-2008 AGEIA Technologies, Inc. All rights reserved.
// Copyright (c) 2001-2004 NovodeX AG. All rights reserved.  


#ifndef PX_PHYSICS_COMMON_TMPMEM
#define PX_PHYSICS_COMMON_TMPMEM

#include "CmPhysXCommon.h"
#include "PsAllocator.h"

namespace physx
{
namespace Cm
{
	// dsequeira: we should be able to use PX_ALLOCA or Ps::InlineArray for this, but both seem slightly flawed:
	//
	//	PX_ALLOCA has non-configurable fallback threshold and uses _alloca, which means the allocation is necessarily
	//	function-scope rather than block-scope (sometimes useful, mostly not.)
	//
	//	Ps::InlineArray touches all memory on resize (a general flaw in the array class which badly needs fixing)
	//
	//	Todo: fix both the above issues.

	template<typename T, PxU32 stackLimit>
	class TmpMem
	{

	public:
		PX_FORCE_INLINE TmpMem(PxU32 size):
		  mPtr(size<=stackLimit?mStackBuf : reinterpret_cast<T*>(PX_ALLOC(size*sizeof(T), "char")))
		{
		}

		PX_FORCE_INLINE ~TmpMem()
		{
			if(mPtr!=mStackBuf)
				PX_FREE(mPtr);
		}

		PX_FORCE_INLINE T& operator*() const
	    {
			return *mPtr;
		}

		PX_FORCE_INLINE T* operator->() const
		{
			return mPtr;
		}
	
		PX_FORCE_INLINE T& operator[](PxU32 index)
		{
			return mPtr[index];
		}

		T* getBase()
		{
			return mPtr;
		}

	private:
		T mStackBuf[stackLimit];
		T* mPtr;
	};
}

}

#endif
