//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of NVIDIA CORPORATION nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Copyright (c) 2008-2021 NVIDIA Corporation. All rights reserved.
// Copyright (c) 2004-2008 AGEIA Technologies, Inc. All rights reserved.
// Copyright (c) 2001-2004 NovodeX AG. All rights reserved.  


#ifndef PX_CONVEXPOLYGONSBUILDER_H
#define PX_CONVEXPOLYGONSBUILDER_H

#include "ConvexHullBuilder.h"

namespace physx
{
	//////////////////////////////////////////////////////////////////////////
	// extended convex hull builder for a case where we build polygons from input triangles
	class ConvexPolygonsBuilder : public ConvexHullBuilder
	{
	public:
		ConvexPolygonsBuilder(Gu::ConvexHullData* hull, const bool buildGRBData);
		~ConvexPolygonsBuilder();

		bool	computeHullPolygons(const PxU32& nbVerts,const PxVec3* verts, const PxU32& nbTriangles, const PxU32* triangles);

		PX_INLINE	PxU32						getNbFaces()const	{ return mNbHullFaces; }
		PX_INLINE	const HullTriangleData*		getFaces()	const	{ return mFaces; }


	private:
		bool	createPolygonData();
		bool	createTrianglesFromPolygons();
		
	private:
		PxU32						mNbHullFaces;	//!< Number of faces in the convex hull
		HullTriangleData*			mFaces;			//!< Triangles.

	};
}

#endif	// PX_CONVEXHULLBUILDER_H

