/****************************************************************************
** Meta object code from reading C++ file 'qjoint_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtquick3dphysics/src/quick3dphysics/joints/qjoint_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qjoint_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN13QPhysicsJointE_t {};
} // unnamed namespace

template <> constexpr inline auto QPhysicsJoint::qt_create_metaobjectdata<qt_meta_tag_ZN13QPhysicsJointE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QPhysicsJoint",
        "QML.Element",
        "PhysicsJoint",
        "QML.Creatable",
        "false",
        "QML.UncreatableReason",
        "abstract interface",
        "bodyAChanged",
        "",
        "bodyBChanged",
        "positionAChanged",
        "positionBChanged",
        "orientationAChanged",
        "orientationBChanged",
        "bodyA",
        "QAbstractPhysicsBody*",
        "bodyB",
        "positionA",
        "QVector3D",
        "positionB",
        "orientationA",
        "QQuaternion",
        "orientationB"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'bodyAChanged'
        QtMocHelpers::RevisionedSignalData<void()>(7, 8, QMC::AccessPublic, 0x60c, QMetaType::Void),
        // Signal 'bodyBChanged'
        QtMocHelpers::RevisionedSignalData<void()>(9, 8, QMC::AccessPublic, 0x60c, QMetaType::Void),
        // Signal 'positionAChanged'
        QtMocHelpers::RevisionedSignalData<void()>(10, 8, QMC::AccessPublic, 0x60c, QMetaType::Void),
        // Signal 'positionBChanged'
        QtMocHelpers::RevisionedSignalData<void()>(11, 8, QMC::AccessPublic, 0x60c, QMetaType::Void),
        // Signal 'orientationAChanged'
        QtMocHelpers::RevisionedSignalData<void()>(12, 8, QMC::AccessPublic, 0x60c, QMetaType::Void),
        // Signal 'orientationBChanged'
        QtMocHelpers::RevisionedSignalData<void()>(13, 8, QMC::AccessPublic, 0x60c, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'bodyA'
        QtMocHelpers::PropertyData<QAbstractPhysicsBody*>(14, 0x80000000 | 15, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet | QMC::Final, 0, 0x60c),
        // property 'bodyB'
        QtMocHelpers::PropertyData<QAbstractPhysicsBody*>(16, 0x80000000 | 15, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet | QMC::Final, 1, 0x60c),
        // property 'positionA'
        QtMocHelpers::PropertyData<QVector3D>(17, 0x80000000 | 18, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet | QMC::Final, 2, 0x60c),
        // property 'positionB'
        QtMocHelpers::PropertyData<QVector3D>(19, 0x80000000 | 18, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet | QMC::Final, 3, 0x60c),
        // property 'orientationA'
        QtMocHelpers::PropertyData<QQuaternion>(20, 0x80000000 | 21, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet | QMC::Final, 4, 0x60c),
        // property 'orientationB'
        QtMocHelpers::PropertyData<QQuaternion>(22, 0x80000000 | 21, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet | QMC::Final, 5, 0x60c),
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
            {    5,    6 },
    });
    return QtMocHelpers::metaObjectData<QPhysicsJoint, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QPhysicsJoint::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QPhysicsJointE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QPhysicsJointE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13QPhysicsJointE_t>.metaTypes,
    nullptr
} };

void QPhysicsJoint::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QPhysicsJoint *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->bodyAChanged(); break;
        case 1: _t->bodyBChanged(); break;
        case 2: _t->positionAChanged(); break;
        case 3: _t->positionBChanged(); break;
        case 4: _t->orientationAChanged(); break;
        case 5: _t->orientationBChanged(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QPhysicsJoint::*)()>(_a, &QPhysicsJoint::bodyAChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsJoint::*)()>(_a, &QPhysicsJoint::bodyBChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsJoint::*)()>(_a, &QPhysicsJoint::positionAChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsJoint::*)()>(_a, &QPhysicsJoint::positionBChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsJoint::*)()>(_a, &QPhysicsJoint::orientationAChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsJoint::*)()>(_a, &QPhysicsJoint::orientationBChanged, 5))
            return;
    }
    if (_c == QMetaObject::RegisterPropertyMetaType) {
        switch (_id) {
        default: *reinterpret_cast<int*>(_a[0]) = -1; break;
        case 1:
        case 0:
            *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAbstractPhysicsBody* >(); break;
        }
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QAbstractPhysicsBody**>(_v) = _t->bodyA(); break;
        case 1: *reinterpret_cast<QAbstractPhysicsBody**>(_v) = _t->bodyB(); break;
        case 2: *reinterpret_cast<QVector3D*>(_v) = _t->positionA(); break;
        case 3: *reinterpret_cast<QVector3D*>(_v) = _t->positionB(); break;
        case 4: *reinterpret_cast<QQuaternion*>(_v) = _t->orientationA(); break;
        case 5: *reinterpret_cast<QQuaternion*>(_v) = _t->orientationB(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setBodyA(*reinterpret_cast<QAbstractPhysicsBody**>(_v)); break;
        case 1: _t->setBodyB(*reinterpret_cast<QAbstractPhysicsBody**>(_v)); break;
        case 2: _t->setPositionA(*reinterpret_cast<QVector3D*>(_v)); break;
        case 3: _t->setPositionB(*reinterpret_cast<QVector3D*>(_v)); break;
        case 4: _t->setOrientationA(*reinterpret_cast<QQuaternion*>(_v)); break;
        case 5: _t->setOrientationB(*reinterpret_cast<QQuaternion*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QPhysicsJoint::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QPhysicsJoint::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QPhysicsJointE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QPhysicsJoint::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 6)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 6;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 6)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 6;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 6;
    }
    return _id;
}

// SIGNAL 0
void QPhysicsJoint::bodyAChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QPhysicsJoint::bodyBChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QPhysicsJoint::positionAChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QPhysicsJoint::positionBChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QPhysicsJoint::orientationAChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QPhysicsJoint::orientationBChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}
QT_WARNING_POP
