//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of NVIDIA CORPORATION nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Copyright (c) 2008-2021 NVIDIA Corporation. All rights reserved.
// Copyright (c) 2004-2008 AGEIA Technologies, Inc. All rights reserved.
// Copyright (c) 2001-2004 NovodeX AG. All rights reserved.

#ifndef PXFOUNDATION_PXBITANDDATA_H
#define PXFOUNDATION_PXBITANDDATA_H

#include "foundation/Px.h"

/** \addtogroup foundation
  @{
*/
#if !PX_DOXYGEN
namespace physx
{
#endif

template <typename storageType, storageType bitMask>
class PxBitAndDataT
{
  public:
	PX_FORCE_INLINE PxBitAndDataT(const PxEMPTY)
	{
	}
	PX_FORCE_INLINE PxBitAndDataT() : mData(0)
	{
	}
	PX_FORCE_INLINE PxBitAndDataT(storageType data, bool bit = false)
	{
		mData = bit ? storageType(data | bitMask) : data;
	}

	PX_CUDA_CALLABLE PX_FORCE_INLINE operator storageType() const
	{
		return storageType(mData & ~bitMask);
	}
	PX_CUDA_CALLABLE PX_FORCE_INLINE void setBit()
	{
		mData |= bitMask;
	}
	PX_CUDA_CALLABLE PX_FORCE_INLINE void clearBit()
	{
		mData &= ~bitMask;
	}
	PX_CUDA_CALLABLE PX_FORCE_INLINE storageType isBitSet() const
	{
		return storageType(mData & bitMask);
	}

  protected:
	storageType mData;
};
typedef PxBitAndDataT<unsigned char, 0x80> PxBitAndByte;
typedef PxBitAndDataT<unsigned short, 0x8000> PxBitAndWord;
typedef PxBitAndDataT<unsigned int, 0x80000000> PxBitAndDword;

#if !PX_DOXYGEN
} // namespace physx
#endif

/** @} */
#endif // PXFOUNDATION_PXBITANDDATA_H
