/****************************************************************************
** Generated QML type registration code
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <QtQml/qqml.h>
#include <QtQml/qqmlmoduleregistration.h>

#if __has_include(<private/qabstractcollisionshape_p.h>)
#  include <private/qabstractcollisionshape_p.h>
#endif
#if __has_include(<private/qabstractphysicsbody_p.h>)
#  include <private/qabstractphysicsbody_p.h>
#endif
#if __has_include(<private/qabstractphysicsnode_p.h>)
#  include <private/qabstractphysicsnode_p.h>
#endif
#if __has_include(<private/qboxshape_p.h>)
#  include <private/qboxshape_p.h>
#endif
#if __has_include(<private/qcapsuleshape_p.h>)
#  include <private/qcapsuleshape_p.h>
#endif
#if __has_include(<private/qcharactercontroller_p.h>)
#  include <private/qcharactercontroller_p.h>
#endif
#if __has_include(<private/qconvexmeshshape_p.h>)
#  include <private/qconvexmeshshape_p.h>
#endif
#if __has_include(<private/qdistancejoint_p.h>)
#  include <private/qdistancejoint_p.h>
#endif
#if __has_include(<private/qdynamicrigidbody_p.h>)
#  include <private/qdynamicrigidbody_p.h>
#endif
#if __has_include(<private/qfixedjoint_p.h>)
#  include <private/qfixedjoint_p.h>
#endif
#if __has_include(<private/qheightfieldshape_p.h>)
#  include <private/qheightfieldshape_p.h>
#endif
#if __has_include(<private/qjoint_p.h>)
#  include <private/qjoint_p.h>
#endif
#if __has_include(<private/qmeshshape_p.h>)
#  include <private/qmeshshape_p.h>
#endif
#if __has_include(<private/qphysicsmaterial_p.h>)
#  include <private/qphysicsmaterial_p.h>
#endif
#if __has_include(<private/qphysicsworld_p.h>)
#  include <private/qphysicsworld_p.h>
#endif
#if __has_include(<private/qplaneshape_p.h>)
#  include <private/qplaneshape_p.h>
#endif
#if __has_include(<private/qprismaticjoint_p.h>)
#  include <private/qprismaticjoint_p.h>
#endif
#if __has_include(<private/qrevolutejoint_p.h>)
#  include <private/qrevolutejoint_p.h>
#endif
#if __has_include(<private/qsphereshape_p.h>)
#  include <private/qsphereshape_p.h>
#endif
#if __has_include(<private/qsphericaljoint_p.h>)
#  include <private/qsphericaljoint_p.h>
#endif
#if __has_include(<private/qstaticrigidbody_p.h>)
#  include <private/qstaticrigidbody_p.h>
#endif
#if __has_include(<private/qtrianglemeshshape_p.h>)
#  include <private/qtrianglemeshshape_p.h>
#endif
#if __has_include(<private/qtriggerbody_p.h>)
#  include <private/qtriggerbody_p.h>
#endif


#if !defined(QT_STATIC)
#define Q_QMLTYPE_EXPORT Q_DECL_EXPORT
#else
#define Q_QMLTYPE_EXPORT
#endif
Q_QMLTYPE_EXPORT void qml_register_types_QtQuick3D_Physics()
{
    qmlRegisterModule("QtQuick3D.Physics", 6, 0);
    QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
    qmlRegisterTypesAndRevisions<QAbstractCollisionShape>("QtQuick3D.Physics", 6);
    qmlRegisterAnonymousTypesAndRevisions<QQuick3DNode>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QAbstractPhysicsBody>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QAbstractPhysicsNode>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QBoxShape>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QCapsuleShape>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QCharacterController>("QtQuick3D.Physics", 6);
    qmlRegisterEnum<QCharacterController::Collisions>("QCharacterController::Collisions");
    qmlRegisterEnum<QCharacterController::Collision>("QCharacterController::Collision");
    qmlRegisterTypesAndRevisions<QConvexMeshShape>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QDistanceJoint>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QDynamicRigidBody>("QtQuick3D.Physics", 6);
    qmlRegisterEnum<QDynamicRigidBody::MassMode>("QDynamicRigidBody::MassMode");
    qmlRegisterEnum<QDynamicRigidBody::AxisLock>("QDynamicRigidBody::AxisLock");
    qmlRegisterTypesAndRevisions<QFixedJoint>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QHeightFieldShape>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QMeshShape>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QPhysicsJoint>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QPhysicsMaterial>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QPhysicsWorld>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QPlaneShape>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QPrismaticJoint>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QRevoluteJoint>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QSphereShape>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QSphericalJoint>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QStaticRigidBody>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QTriangleMeshShape>("QtQuick3D.Physics", 6);
    qmlRegisterTypesAndRevisions<QTriggerBody>("QtQuick3D.Physics", 6);
    QT_WARNING_POP
    qmlRegisterModule("QtQuick3D.Physics", 6, 12);
}

static const QQmlModuleRegistration qtQuick3DPhysicsRegistration("QtQuick3D.Physics", qml_register_types_QtQuick3D_Physics);
