/****************************************************************************
** Meta object code from reading C++ file 'qcapsulegeometry_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtquick3dphysics/src/helpers/qcapsulegeometry_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qcapsulegeometry_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN22CapsuleGeometryPhysicsE_t {};
} // unnamed namespace

template <> constexpr inline auto CapsuleGeometryPhysics::qt_create_metaobjectdata<qt_meta_tag_ZN22CapsuleGeometryPhysicsE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "CapsuleGeometryPhysics",
        "QML.Element",
        "CapsuleGeometry",
        "enableNormalsChanged",
        "",
        "enableUVChanged",
        "longitudesChanged",
        "latitudesChanged",
        "ringsChanged",
        "heightChanged",
        "diameterChanged",
        "enableNormals",
        "enableUV",
        "longitudes",
        "latitudes",
        "rings",
        "height",
        "diameter"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'enableNormalsChanged'
        QtMocHelpers::SignalData<void()>(3, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'enableUVChanged'
        QtMocHelpers::SignalData<void()>(5, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'longitudesChanged'
        QtMocHelpers::SignalData<void()>(6, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'latitudesChanged'
        QtMocHelpers::SignalData<void()>(7, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'ringsChanged'
        QtMocHelpers::SignalData<void()>(8, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'heightChanged'
        QtMocHelpers::SignalData<void()>(9, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'diameterChanged'
        QtMocHelpers::SignalData<void()>(10, 4, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'enableNormals'
        QtMocHelpers::PropertyData<bool>(11, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 0),
        // property 'enableUV'
        QtMocHelpers::PropertyData<bool>(12, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 1),
        // property 'longitudes'
        QtMocHelpers::PropertyData<int>(13, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 2),
        // property 'latitudes'
        QtMocHelpers::PropertyData<int>(14, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 3),
        // property 'rings'
        QtMocHelpers::PropertyData<int>(15, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 4),
        // property 'height'
        QtMocHelpers::PropertyData<float>(16, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 5),
        // property 'diameter'
        QtMocHelpers::PropertyData<float>(17, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 6),
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<CapsuleGeometryPhysics, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject CapsuleGeometryPhysics::staticMetaObject = { {
    QMetaObject::SuperData::link<QQuick3DGeometry::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN22CapsuleGeometryPhysicsE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN22CapsuleGeometryPhysicsE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN22CapsuleGeometryPhysicsE_t>.metaTypes,
    nullptr
} };

void CapsuleGeometryPhysics::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<CapsuleGeometryPhysics *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->enableNormalsChanged(); break;
        case 1: _t->enableUVChanged(); break;
        case 2: _t->longitudesChanged(); break;
        case 3: _t->latitudesChanged(); break;
        case 4: _t->ringsChanged(); break;
        case 5: _t->heightChanged(); break;
        case 6: _t->diameterChanged(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (CapsuleGeometryPhysics::*)()>(_a, &CapsuleGeometryPhysics::enableNormalsChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (CapsuleGeometryPhysics::*)()>(_a, &CapsuleGeometryPhysics::enableUVChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (CapsuleGeometryPhysics::*)()>(_a, &CapsuleGeometryPhysics::longitudesChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (CapsuleGeometryPhysics::*)()>(_a, &CapsuleGeometryPhysics::latitudesChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (CapsuleGeometryPhysics::*)()>(_a, &CapsuleGeometryPhysics::ringsChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (CapsuleGeometryPhysics::*)()>(_a, &CapsuleGeometryPhysics::heightChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (CapsuleGeometryPhysics::*)()>(_a, &CapsuleGeometryPhysics::diameterChanged, 6))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<bool*>(_v) = _t->enableNormals(); break;
        case 1: *reinterpret_cast<bool*>(_v) = _t->enableUV(); break;
        case 2: *reinterpret_cast<int*>(_v) = _t->longitudes(); break;
        case 3: *reinterpret_cast<int*>(_v) = _t->latitudes(); break;
        case 4: *reinterpret_cast<int*>(_v) = _t->rings(); break;
        case 5: *reinterpret_cast<float*>(_v) = _t->height(); break;
        case 6: *reinterpret_cast<float*>(_v) = _t->diameter(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setEnableNormals(*reinterpret_cast<bool*>(_v)); break;
        case 1: _t->setEnableUV(*reinterpret_cast<bool*>(_v)); break;
        case 2: _t->setLongitudes(*reinterpret_cast<int*>(_v)); break;
        case 3: _t->setLatitudes(*reinterpret_cast<int*>(_v)); break;
        case 4: _t->setRings(*reinterpret_cast<int*>(_v)); break;
        case 5: _t->setHeight(*reinterpret_cast<float*>(_v)); break;
        case 6: _t->setDiameter(*reinterpret_cast<float*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *CapsuleGeometryPhysics::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *CapsuleGeometryPhysics::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN22CapsuleGeometryPhysicsE_t>.strings))
        return static_cast<void*>(this);
    return QQuick3DGeometry::qt_metacast(_clname);
}

int CapsuleGeometryPhysics::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QQuick3DGeometry::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 7)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 7)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 7;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    return _id;
}

// SIGNAL 0
void CapsuleGeometryPhysics::enableNormalsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void CapsuleGeometryPhysics::enableUVChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void CapsuleGeometryPhysics::longitudesChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void CapsuleGeometryPhysics::latitudesChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void CapsuleGeometryPhysics::ringsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void CapsuleGeometryPhysics::heightChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void CapsuleGeometryPhysics::diameterChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}
QT_WARNING_POP
