/****************************************************************************
** Meta object code from reading C++ file 'qcharactercontroller_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtquick3dphysics/src/quick3dphysics/qcharactercontroller_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qcharactercontroller_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN20QCharacterControllerE_t {};
} // unnamed namespace

template <> constexpr inline auto QCharacterController::qt_create_metaobjectdata<qt_meta_tag_ZN20QCharacterControllerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QCharacterController",
        "QML.Element",
        "CharacterController",
        "movementChanged",
        "",
        "gravityChanged",
        "midAirControlChanged",
        "collisionsChanged",
        "enableShapeHitCallbackChanged",
        "shapeHit",
        "QAbstractPhysicsNode*",
        "body",
        "QVector3D",
        "position",
        "impulse",
        "normal",
        "teleport",
        "movement",
        "gravity",
        "midAirControl",
        "collisions",
        "Collisions",
        "enableShapeHitCallback",
        "Collision",
        "None",
        "Side",
        "Up",
        "Down"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'movementChanged'
        QtMocHelpers::SignalData<void()>(3, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'gravityChanged'
        QtMocHelpers::SignalData<void()>(5, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'midAirControlChanged'
        QtMocHelpers::SignalData<void()>(6, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'collisionsChanged'
        QtMocHelpers::SignalData<void()>(7, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'enableShapeHitCallbackChanged'
        QtMocHelpers::SignalData<void()>(8, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'shapeHit'
        QtMocHelpers::SignalData<void(QAbstractPhysicsNode *, const QVector3D &, const QVector3D &, const QVector3D &)>(9, 4, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 }, { 0x80000000 | 12, 13 }, { 0x80000000 | 12, 14 }, { 0x80000000 | 12, 15 },
        }}),
        // Method 'teleport'
        QtMocHelpers::MethodData<void(const QVector3D &)>(16, 4, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 13 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'movement'
        QtMocHelpers::PropertyData<QVector3D>(17, 0x80000000 | 12, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 0),
        // property 'gravity'
        QtMocHelpers::PropertyData<QVector3D>(18, 0x80000000 | 12, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 1),
        // property 'midAirControl'
        QtMocHelpers::PropertyData<bool>(19, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 2),
        // property 'collisions'
        QtMocHelpers::PropertyData<Collisions>(20, 0x80000000 | 21, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 3),
        // property 'enableShapeHitCallback'
        QtMocHelpers::PropertyData<bool>(22, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 4),
    };
    QtMocHelpers::UintData qt_enums {
        // flag 'Collisions'
        QtMocHelpers::EnumData<Collisions>(21, 23, QMC::EnumIsFlag | QMC::EnumIsScoped).add({
            {   24, Collision::None },
            {   25, Collision::Side },
            {   26, Collision::Up },
            {   27, Collision::Down },
        }),
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<QCharacterController, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QCharacterController::staticMetaObject = { {
    QMetaObject::SuperData::link<QAbstractPhysicsBody::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QCharacterControllerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QCharacterControllerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN20QCharacterControllerE_t>.metaTypes,
    nullptr
} };

void QCharacterController::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QCharacterController *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->movementChanged(); break;
        case 1: _t->gravityChanged(); break;
        case 2: _t->midAirControlChanged(); break;
        case 3: _t->collisionsChanged(); break;
        case 4: _t->enableShapeHitCallbackChanged(); break;
        case 5: _t->shapeHit((*reinterpret_cast<std::add_pointer_t<QAbstractPhysicsNode*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QVector3D>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QVector3D>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<QVector3D>>(_a[4]))); break;
        case 6: _t->teleport((*reinterpret_cast<std::add_pointer_t<QVector3D>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 5:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QAbstractPhysicsNode* >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QCharacterController::*)()>(_a, &QCharacterController::movementChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QCharacterController::*)()>(_a, &QCharacterController::gravityChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QCharacterController::*)()>(_a, &QCharacterController::midAirControlChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QCharacterController::*)()>(_a, &QCharacterController::collisionsChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QCharacterController::*)()>(_a, &QCharacterController::enableShapeHitCallbackChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QCharacterController::*)(QAbstractPhysicsNode * , const QVector3D & , const QVector3D & , const QVector3D & )>(_a, &QCharacterController::shapeHit, 5))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QVector3D*>(_v) = _t->movement(); break;
        case 1: *reinterpret_cast<QVector3D*>(_v) = _t->gravity(); break;
        case 2: *reinterpret_cast<bool*>(_v) = _t->midAirControl(); break;
        case 3: QtMocHelpers::assignFlags<Collisions>(_v, _t->collisions()); break;
        case 4: *reinterpret_cast<bool*>(_v) = _t->enableShapeHitCallback(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setMovement(*reinterpret_cast<QVector3D*>(_v)); break;
        case 1: _t->setGravity(*reinterpret_cast<QVector3D*>(_v)); break;
        case 2: _t->setMidAirControl(*reinterpret_cast<bool*>(_v)); break;
        case 4: _t->setEnableShapeHitCallback(*reinterpret_cast<bool*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QCharacterController::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QCharacterController::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QCharacterControllerE_t>.strings))
        return static_cast<void*>(this);
    return QAbstractPhysicsBody::qt_metacast(_clname);
}

int QCharacterController::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAbstractPhysicsBody::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 7)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 7)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 5;
    }
    return _id;
}

// SIGNAL 0
void QCharacterController::movementChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QCharacterController::gravityChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QCharacterController::midAirControlChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QCharacterController::collisionsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QCharacterController::enableShapeHitCallbackChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QCharacterController::shapeHit(QAbstractPhysicsNode * _t1, const QVector3D & _t2, const QVector3D & _t3, const QVector3D & _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1, _t2, _t3, _t4);
}
QT_WARNING_POP
