/****************************************************************************
** Meta object code from reading C++ file 'qheightfieldshape_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtquick3dphysics/src/quick3dphysics/qheightfieldshape_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qheightfieldshape_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN17QHeightFieldShapeE_t {};
} // unnamed namespace

template <> constexpr inline auto QHeightFieldShape::qt_create_metaobjectdata<qt_meta_tag_ZN17QHeightFieldShapeE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QHeightFieldShape",
        "QML.Element",
        "HeightFieldShape",
        "sourceChanged",
        "",
        "extentsChanged",
        "imageChanged",
        "imageDestroyed",
        "image",
        "imageGeometryChanged",
        "extents",
        "QVector3D",
        "source",
        "QUrl",
        "QQuickImage*"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'sourceChanged'
        QtMocHelpers::RevisionedSignalData<void()>(3, 4, QMC::AccessPublic, 0x605, QMetaType::Void),
        // Signal 'extentsChanged'
        QtMocHelpers::SignalData<void()>(5, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'imageChanged'
        QtMocHelpers::RevisionedSignalData<void()>(6, 4, QMC::AccessPublic, 0x607, QMetaType::Void),
        // Slot 'imageDestroyed'
        QtMocHelpers::SlotData<void(QObject *)>(7, 4, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QObjectStar, 8 },
        }}),
        // Slot 'imageGeometryChanged'
        QtMocHelpers::SlotData<void()>(9, 4, QMC::AccessPrivate, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'extents'
        QtMocHelpers::PropertyData<QVector3D>(10, 0x80000000 | 11, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 1),
        // property 'source'
        QtMocHelpers::PropertyData<QUrl>(12, 0x80000000 | 13, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 0, 0x605),
        // property 'image'
        QtMocHelpers::PropertyData<QQuickImage*>(8, 0x80000000 | 14, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 2, 0x607),
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<QHeightFieldShape, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QHeightFieldShape::staticMetaObject = { {
    QMetaObject::SuperData::link<QAbstractCollisionShape::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QHeightFieldShapeE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QHeightFieldShapeE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN17QHeightFieldShapeE_t>.metaTypes,
    nullptr
} };

void QHeightFieldShape::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QHeightFieldShape *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->sourceChanged(); break;
        case 1: _t->extentsChanged(); break;
        case 2: _t->imageChanged(); break;
        case 3: _t->imageDestroyed((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1]))); break;
        case 4: _t->imageGeometryChanged(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QHeightFieldShape::*)()>(_a, &QHeightFieldShape::sourceChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QHeightFieldShape::*)()>(_a, &QHeightFieldShape::extentsChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QHeightFieldShape::*)()>(_a, &QHeightFieldShape::imageChanged, 2))
            return;
    }
    if (_c == QMetaObject::RegisterPropertyMetaType) {
        switch (_id) {
        default: *reinterpret_cast<int*>(_a[0]) = -1; break;
        case 2:
            *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QQuickImage* >(); break;
        }
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QVector3D*>(_v) = _t->extents(); break;
        case 1: *reinterpret_cast<QUrl*>(_v) = _t->source(); break;
        case 2: *reinterpret_cast<QQuickImage**>(_v) = _t->image(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setExtents(*reinterpret_cast<QVector3D*>(_v)); break;
        case 1: _t->setSource(*reinterpret_cast<QUrl*>(_v)); break;
        case 2: _t->setImage(*reinterpret_cast<QQuickImage**>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QHeightFieldShape::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QHeightFieldShape::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QHeightFieldShapeE_t>.strings))
        return static_cast<void*>(this);
    return QAbstractCollisionShape::qt_metacast(_clname);
}

int QHeightFieldShape::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAbstractCollisionShape::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 5)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 5;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 5)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 5;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 3;
    }
    return _id;
}

// SIGNAL 0
void QHeightFieldShape::sourceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QHeightFieldShape::extentsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QHeightFieldShape::imageChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}
QT_WARNING_POP
