/****************************************************************************
** Meta object code from reading C++ file 'qphysicsworld_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtquick3dphysics/src/quick3dphysics/qphysicsworld_p.h"
#include <QtGui/qtextcursor.h>
#include <QtNetwork/QSslError>
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qphysicsworld_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN13QPhysicsWorldE_t {};
} // unnamed namespace

template <> constexpr inline auto QPhysicsWorld::qt_create_metaobjectdata<qt_meta_tag_ZN13QPhysicsWorldE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QPhysicsWorld",
        "QML.Element",
        "PhysicsWorld",
        "gravityChanged",
        "",
        "QVector3D",
        "gravity",
        "runningChanged",
        "running",
        "enableCCDChanged",
        "enableCCD",
        "forceDebugDrawChanged",
        "forceDebugDraw",
        "typicalLengthChanged",
        "typicalLength",
        "typicalSpeedChanged",
        "typicalSpeed",
        "defaultDensityChanged",
        "defaultDensity",
        "viewportChanged",
        "QQuick3DNode*",
        "viewport",
        "minimumTimestepChanged",
        "minimumTimestep",
        "maximumTimestepChanged",
        "maxTimestep",
        "frameDone",
        "timestep",
        "sceneChanged",
        "numThreadsChanged",
        "reportKinematicKinematicCollisionsChanged",
        "reportStaticKinematicCollisionsChanged",
        "setGravity",
        "setRunning",
        "setForceDebugDraw",
        "setEnableCCD",
        "setTypicalLength",
        "setTypicalSpeed",
        "setDefaultDensity",
        "setViewport",
        "setMinimumTimestep",
        "minTimestep",
        "setMaximumTimestep",
        "setScene",
        "newScene",
        "setNumThreads",
        "newNumThreads",
        "maximumTimestep",
        "scene",
        "numThreads",
        "reportKinematicKinematicCollisions",
        "reportStaticKinematicCollisions"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'gravityChanged'
        QtMocHelpers::SignalData<void(QVector3D)>(3, 4, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 6 },
        }}),
        // Signal 'runningChanged'
        QtMocHelpers::SignalData<void(bool)>(7, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 8 },
        }}),
        // Signal 'enableCCDChanged'
        QtMocHelpers::SignalData<void(bool)>(9, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 10 },
        }}),
        // Signal 'forceDebugDrawChanged'
        QtMocHelpers::SignalData<void(bool)>(11, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 12 },
        }}),
        // Signal 'typicalLengthChanged'
        QtMocHelpers::SignalData<void(float)>(13, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 14 },
        }}),
        // Signal 'typicalSpeedChanged'
        QtMocHelpers::SignalData<void(float)>(15, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 16 },
        }}),
        // Signal 'defaultDensityChanged'
        QtMocHelpers::SignalData<void(float)>(17, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 18 },
        }}),
        // Signal 'viewportChanged'
        QtMocHelpers::RevisionedSignalData<void(QQuick3DNode *)>(19, 4, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { 0x80000000 | 20, 21 },
        }}),
        // Signal 'minimumTimestepChanged'
        QtMocHelpers::RevisionedSignalData<void(float)>(22, 4, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { QMetaType::Float, 23 },
        }}),
        // Signal 'maximumTimestepChanged'
        QtMocHelpers::RevisionedSignalData<void(float)>(24, 4, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { QMetaType::Float, 25 },
        }}),
        // Signal 'frameDone'
        QtMocHelpers::RevisionedSignalData<void(float)>(26, 4, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { QMetaType::Float, 27 },
        }}),
        // Signal 'sceneChanged'
        QtMocHelpers::RevisionedSignalData<void()>(28, 4, QMC::AccessPublic, 0x605, QMetaType::Void),
        // Signal 'numThreadsChanged'
        QtMocHelpers::RevisionedSignalData<void()>(29, 4, QMC::AccessPublic, 0x607, QMetaType::Void),
        // Signal 'reportKinematicKinematicCollisionsChanged'
        QtMocHelpers::RevisionedSignalData<void()>(30, 4, QMC::AccessPublic, 0x607, QMetaType::Void),
        // Signal 'reportStaticKinematicCollisionsChanged'
        QtMocHelpers::RevisionedSignalData<void()>(31, 4, QMC::AccessPublic, 0x607, QMetaType::Void),
        // Slot 'setGravity'
        QtMocHelpers::SlotData<void(QVector3D)>(32, 4, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 6 },
        }}),
        // Slot 'setRunning'
        QtMocHelpers::SlotData<void(bool)>(33, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 8 },
        }}),
        // Slot 'setForceDebugDraw'
        QtMocHelpers::SlotData<void(bool)>(34, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 12 },
        }}),
        // Slot 'setEnableCCD'
        QtMocHelpers::SlotData<void(bool)>(35, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 10 },
        }}),
        // Slot 'setTypicalLength'
        QtMocHelpers::SlotData<void(float)>(36, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 14 },
        }}),
        // Slot 'setTypicalSpeed'
        QtMocHelpers::SlotData<void(float)>(37, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 16 },
        }}),
        // Slot 'setDefaultDensity'
        QtMocHelpers::SlotData<void(float)>(38, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 18 },
        }}),
        // Slot 'setViewport'
        QtMocHelpers::RevisionedSlotData<void(QQuick3DNode *)>(39, 4, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { 0x80000000 | 20, 21 },
        }}),
        // Slot 'setMinimumTimestep'
        QtMocHelpers::RevisionedSlotData<void(float)>(40, 4, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { QMetaType::Float, 41 },
        }}),
        // Slot 'setMaximumTimestep'
        QtMocHelpers::RevisionedSlotData<void(float)>(42, 4, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { QMetaType::Float, 25 },
        }}),
        // Slot 'setScene'
        QtMocHelpers::RevisionedSlotData<void(QQuick3DNode *)>(43, 4, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { 0x80000000 | 20, 44 },
        }}),
        // Slot 'setNumThreads'
        QtMocHelpers::RevisionedSlotData<void(int)>(45, 4, QMC::AccessPublic, 0x607, QMetaType::Void, {{
            { QMetaType::Int, 46 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'gravity'
        QtMocHelpers::PropertyData<QVector3D>(6, 0x80000000 | 5, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 0),
        // property 'running'
        QtMocHelpers::PropertyData<bool>(8, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 1),
        // property 'forceDebugDraw'
        QtMocHelpers::PropertyData<bool>(12, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 3),
        // property 'enableCCD'
        QtMocHelpers::PropertyData<bool>(10, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 2),
        // property 'typicalLength'
        QtMocHelpers::PropertyData<float>(14, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 4),
        // property 'typicalSpeed'
        QtMocHelpers::PropertyData<float>(16, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 5),
        // property 'defaultDensity'
        QtMocHelpers::PropertyData<float>(18, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 6),
        // property 'viewport'
        QtMocHelpers::PropertyData<QQuick3DNode*>(21, 0x80000000 | 20, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 7, 0x605),
        // property 'minimumTimestep'
        QtMocHelpers::PropertyData<float>(23, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 8, 0x605),
        // property 'maximumTimestep'
        QtMocHelpers::PropertyData<float>(47, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 9, 0x605),
        // property 'scene'
        QtMocHelpers::PropertyData<QQuick3DNode*>(48, 0x80000000 | 20, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 11, 0x605),
        // property 'numThreads'
        QtMocHelpers::PropertyData<int>(49, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 12, 0x607),
        // property 'reportKinematicKinematicCollisions'
        QtMocHelpers::PropertyData<bool>(50, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet | QMC::Final, 13, 0x607),
        // property 'reportStaticKinematicCollisions'
        QtMocHelpers::PropertyData<bool>(51, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet | QMC::Final, 14, 0x607),
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<QPhysicsWorld, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QPhysicsWorld::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QPhysicsWorldE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QPhysicsWorldE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13QPhysicsWorldE_t>.metaTypes,
    nullptr
} };

void QPhysicsWorld::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QPhysicsWorld *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->gravityChanged((*reinterpret_cast<std::add_pointer_t<QVector3D>>(_a[1]))); break;
        case 1: _t->runningChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 2: _t->enableCCDChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 3: _t->forceDebugDrawChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 4: _t->typicalLengthChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 5: _t->typicalSpeedChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 6: _t->defaultDensityChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 7: _t->viewportChanged((*reinterpret_cast<std::add_pointer_t<QQuick3DNode*>>(_a[1]))); break;
        case 8: _t->minimumTimestepChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 9: _t->maximumTimestepChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 10: _t->frameDone((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 11: _t->sceneChanged(); break;
        case 12: _t->numThreadsChanged(); break;
        case 13: _t->reportKinematicKinematicCollisionsChanged(); break;
        case 14: _t->reportStaticKinematicCollisionsChanged(); break;
        case 15: _t->setGravity((*reinterpret_cast<std::add_pointer_t<QVector3D>>(_a[1]))); break;
        case 16: _t->setRunning((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 17: _t->setForceDebugDraw((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 18: _t->setEnableCCD((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 19: _t->setTypicalLength((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 20: _t->setTypicalSpeed((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 21: _t->setDefaultDensity((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 22: _t->setViewport((*reinterpret_cast<std::add_pointer_t<QQuick3DNode*>>(_a[1]))); break;
        case 23: _t->setMinimumTimestep((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 24: _t->setMaximumTimestep((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 25: _t->setScene((*reinterpret_cast<std::add_pointer_t<QQuick3DNode*>>(_a[1]))); break;
        case 26: _t->setNumThreads((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 7:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QQuick3DNode* >(); break;
            }
            break;
        case 22:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QQuick3DNode* >(); break;
            }
            break;
        case 25:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QQuick3DNode* >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(QVector3D )>(_a, &QPhysicsWorld::gravityChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(bool )>(_a, &QPhysicsWorld::runningChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(bool )>(_a, &QPhysicsWorld::enableCCDChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(bool )>(_a, &QPhysicsWorld::forceDebugDrawChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(float )>(_a, &QPhysicsWorld::typicalLengthChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(float )>(_a, &QPhysicsWorld::typicalSpeedChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(float )>(_a, &QPhysicsWorld::defaultDensityChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(QQuick3DNode * )>(_a, &QPhysicsWorld::viewportChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(float )>(_a, &QPhysicsWorld::minimumTimestepChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(float )>(_a, &QPhysicsWorld::maximumTimestepChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)(float )>(_a, &QPhysicsWorld::frameDone, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)()>(_a, &QPhysicsWorld::sceneChanged, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)()>(_a, &QPhysicsWorld::numThreadsChanged, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)()>(_a, &QPhysicsWorld::reportKinematicKinematicCollisionsChanged, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPhysicsWorld::*)()>(_a, &QPhysicsWorld::reportStaticKinematicCollisionsChanged, 14))
            return;
    }
    if (_c == QMetaObject::RegisterPropertyMetaType) {
        switch (_id) {
        default: *reinterpret_cast<int*>(_a[0]) = -1; break;
        case 10:
        case 7:
            *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QQuick3DNode* >(); break;
        }
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QVector3D*>(_v) = _t->gravity(); break;
        case 1: *reinterpret_cast<bool*>(_v) = _t->running(); break;
        case 2: *reinterpret_cast<bool*>(_v) = _t->forceDebugDraw(); break;
        case 3: *reinterpret_cast<bool*>(_v) = _t->enableCCD(); break;
        case 4: *reinterpret_cast<float*>(_v) = _t->typicalLength(); break;
        case 5: *reinterpret_cast<float*>(_v) = _t->typicalSpeed(); break;
        case 6: *reinterpret_cast<float*>(_v) = _t->defaultDensity(); break;
        case 7: *reinterpret_cast<QQuick3DNode**>(_v) = _t->viewport(); break;
        case 8: *reinterpret_cast<float*>(_v) = _t->minimumTimestep(); break;
        case 9: *reinterpret_cast<float*>(_v) = _t->maximumTimestep(); break;
        case 10: *reinterpret_cast<QQuick3DNode**>(_v) = _t->scene(); break;
        case 11: *reinterpret_cast<int*>(_v) = _t->numThreads(); break;
        case 12: *reinterpret_cast<bool*>(_v) = _t->reportKinematicKinematicCollisions(); break;
        case 13: *reinterpret_cast<bool*>(_v) = _t->reportStaticKinematicCollisions(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setGravity(*reinterpret_cast<QVector3D*>(_v)); break;
        case 1: _t->setRunning(*reinterpret_cast<bool*>(_v)); break;
        case 2: _t->setForceDebugDraw(*reinterpret_cast<bool*>(_v)); break;
        case 3: _t->setEnableCCD(*reinterpret_cast<bool*>(_v)); break;
        case 4: _t->setTypicalLength(*reinterpret_cast<float*>(_v)); break;
        case 5: _t->setTypicalSpeed(*reinterpret_cast<float*>(_v)); break;
        case 6: _t->setDefaultDensity(*reinterpret_cast<float*>(_v)); break;
        case 7: _t->setViewport(*reinterpret_cast<QQuick3DNode**>(_v)); break;
        case 8: _t->setMinimumTimestep(*reinterpret_cast<float*>(_v)); break;
        case 9: _t->setMaximumTimestep(*reinterpret_cast<float*>(_v)); break;
        case 10: _t->setScene(*reinterpret_cast<QQuick3DNode**>(_v)); break;
        case 11: _t->setNumThreads(*reinterpret_cast<int*>(_v)); break;
        case 12: _t->setReportKinematicKinematicCollisions(*reinterpret_cast<bool*>(_v)); break;
        case 13: _t->setReportStaticKinematicCollisions(*reinterpret_cast<bool*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QPhysicsWorld::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QPhysicsWorld::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QPhysicsWorldE_t>.strings))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "QQmlParserStatus"))
        return static_cast< QQmlParserStatus*>(this);
    if (!strcmp(_clname, "org.qt-project.Qt.QQmlParserStatus"))
        return static_cast< QQmlParserStatus*>(this);
    return QObject::qt_metacast(_clname);
}

int QPhysicsWorld::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 27)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 27;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 27)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 27;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 14;
    }
    return _id;
}

// SIGNAL 0
void QPhysicsWorld::gravityChanged(QVector3D _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void QPhysicsWorld::runningChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void QPhysicsWorld::enableCCDChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void QPhysicsWorld::forceDebugDrawChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void QPhysicsWorld::typicalLengthChanged(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void QPhysicsWorld::typicalSpeedChanged(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void QPhysicsWorld::defaultDensityChanged(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void QPhysicsWorld::viewportChanged(QQuick3DNode * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void QPhysicsWorld::minimumTimestepChanged(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void QPhysicsWorld::maximumTimestepChanged(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}

// SIGNAL 10
void QPhysicsWorld::frameDone(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void QPhysicsWorld::sceneChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 11, nullptr);
}

// SIGNAL 12
void QPhysicsWorld::numThreadsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 12, nullptr);
}

// SIGNAL 13
void QPhysicsWorld::reportKinematicKinematicCollisionsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 13, nullptr);
}

// SIGNAL 14
void QPhysicsWorld::reportStaticKinematicCollisionsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 14, nullptr);
}
QT_WARNING_POP
