/****************************************************************************
** Meta object code from reading C++ file 'qabstractphysicsnode_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtquick3dphysics/src/quick3dphysics/qabstractphysicsnode_p.h"
#include <QtCore/qmetatype.h>
#include <QtCore/QList>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qabstractphysicsnode_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN20QAbstractPhysicsNodeE_t {};
} // unnamed namespace

template <> constexpr inline auto QAbstractPhysicsNode::qt_create_metaobjectdata<qt_meta_tag_ZN20QAbstractPhysicsNodeE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QAbstractPhysicsNode",
        "QML.Element",
        "PhysicsNode",
        "QML.Creatable",
        "false",
        "QML.UncreatableReason",
        "abstract interface",
        "bodyContact",
        "",
        "QAbstractPhysicsNode*",
        "body",
        "QList<QVector3D>",
        "positions",
        "impulses",
        "normals",
        "sendContactReportsChanged",
        "sendContactReports",
        "receiveContactReportsChanged",
        "receiveContactReports",
        "sendTriggerReportsChanged",
        "sendTriggerReports",
        "receiveTriggerReportsChanged",
        "receiveTriggerReports",
        "enteredTriggerBody",
        "exitedTriggerBody",
        "filterGroupChanged",
        "filterIgnoreGroupsChanged",
        "onShapeDestroyed",
        "object",
        "onShapeNeedsRebuild",
        "collisionShapes",
        "QQmlListProperty<QAbstractCollisionShape>",
        "filterGroup",
        "filterIgnoreGroups"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'bodyContact'
        QtMocHelpers::SignalData<void(QAbstractPhysicsNode *, const QVector<QVector3D> &, const QVector<QVector3D> &, const QVector<QVector3D> &)>(7, 8, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 9, 10 }, { 0x80000000 | 11, 12 }, { 0x80000000 | 11, 13 }, { 0x80000000 | 11, 14 },
        }}),
        // Signal 'sendContactReportsChanged'
        QtMocHelpers::SignalData<void(float)>(15, 8, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 16 },
        }}),
        // Signal 'receiveContactReportsChanged'
        QtMocHelpers::SignalData<void(float)>(17, 8, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 18 },
        }}),
        // Signal 'sendTriggerReportsChanged'
        QtMocHelpers::RevisionedSignalData<void(float)>(19, 8, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { QMetaType::Float, 20 },
        }}),
        // Signal 'receiveTriggerReportsChanged'
        QtMocHelpers::RevisionedSignalData<void(float)>(21, 8, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { QMetaType::Float, 22 },
        }}),
        // Signal 'enteredTriggerBody'
        QtMocHelpers::RevisionedSignalData<void(QAbstractPhysicsNode *)>(23, 8, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { 0x80000000 | 9, 10 },
        }}),
        // Signal 'exitedTriggerBody'
        QtMocHelpers::RevisionedSignalData<void(QAbstractPhysicsNode *)>(24, 8, QMC::AccessPublic, 0x605, QMetaType::Void, {{
            { 0x80000000 | 9, 10 },
        }}),
        // Signal 'filterGroupChanged'
        QtMocHelpers::RevisionedSignalData<void()>(25, 8, QMC::AccessPublic, 0x607, QMetaType::Void),
        // Signal 'filterIgnoreGroupsChanged'
        QtMocHelpers::RevisionedSignalData<void()>(26, 8, QMC::AccessPublic, 0x607, QMetaType::Void),
        // Slot 'onShapeDestroyed'
        QtMocHelpers::SlotData<void(QObject *)>(27, 8, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QObjectStar, 28 },
        }}),
        // Slot 'onShapeNeedsRebuild'
        QtMocHelpers::SlotData<void(QObject *)>(29, 8, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QObjectStar, 28 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'collisionShapes'
        QtMocHelpers::PropertyData<QQmlListProperty<QAbstractCollisionShape>>(30, 0x80000000 | 31, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant),
        // property 'sendContactReports'
        QtMocHelpers::PropertyData<bool>(16, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 1),
        // property 'receiveContactReports'
        QtMocHelpers::PropertyData<bool>(18, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 2),
        // property 'sendTriggerReports'
        QtMocHelpers::PropertyData<bool>(20, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 3, 0x605),
        // property 'receiveTriggerReports'
        QtMocHelpers::PropertyData<bool>(22, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 4, 0x605),
        // property 'filterGroup'
        QtMocHelpers::PropertyData<int>(32, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable, 7, 0x607),
        // property 'filterIgnoreGroups'
        QtMocHelpers::PropertyData<int>(33, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 8, 0x607),
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
            {    5,    6 },
    });
    return QtMocHelpers::metaObjectData<QAbstractPhysicsNode, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QAbstractPhysicsNode::staticMetaObject = { {
    QMetaObject::SuperData::link<QQuick3DNode::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QAbstractPhysicsNodeE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QAbstractPhysicsNodeE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN20QAbstractPhysicsNodeE_t>.metaTypes,
    nullptr
} };

void QAbstractPhysicsNode::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QAbstractPhysicsNode *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->bodyContact((*reinterpret_cast<std::add_pointer_t<QAbstractPhysicsNode*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QList<QVector3D>>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QList<QVector3D>>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<QList<QVector3D>>>(_a[4]))); break;
        case 1: _t->sendContactReportsChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 2: _t->receiveContactReportsChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 3: _t->sendTriggerReportsChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 4: _t->receiveTriggerReportsChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 5: _t->enteredTriggerBody((*reinterpret_cast<std::add_pointer_t<QAbstractPhysicsNode*>>(_a[1]))); break;
        case 6: _t->exitedTriggerBody((*reinterpret_cast<std::add_pointer_t<QAbstractPhysicsNode*>>(_a[1]))); break;
        case 7: _t->filterGroupChanged(); break;
        case 8: _t->filterIgnoreGroupsChanged(); break;
        case 9: _t->onShapeDestroyed((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1]))); break;
        case 10: _t->onShapeNeedsRebuild((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 0:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QAbstractPhysicsNode* >(); break;
            }
            break;
        case 5:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QAbstractPhysicsNode* >(); break;
            }
            break;
        case 6:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QAbstractPhysicsNode* >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QAbstractPhysicsNode::*)(QAbstractPhysicsNode * , const QVector<QVector3D> & , const QVector<QVector3D> & , const QVector<QVector3D> & )>(_a, &QAbstractPhysicsNode::bodyContact, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAbstractPhysicsNode::*)(float )>(_a, &QAbstractPhysicsNode::sendContactReportsChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAbstractPhysicsNode::*)(float )>(_a, &QAbstractPhysicsNode::receiveContactReportsChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAbstractPhysicsNode::*)(float )>(_a, &QAbstractPhysicsNode::sendTriggerReportsChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAbstractPhysicsNode::*)(float )>(_a, &QAbstractPhysicsNode::receiveTriggerReportsChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAbstractPhysicsNode::*)(QAbstractPhysicsNode * )>(_a, &QAbstractPhysicsNode::enteredTriggerBody, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAbstractPhysicsNode::*)(QAbstractPhysicsNode * )>(_a, &QAbstractPhysicsNode::exitedTriggerBody, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAbstractPhysicsNode::*)()>(_a, &QAbstractPhysicsNode::filterGroupChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAbstractPhysicsNode::*)()>(_a, &QAbstractPhysicsNode::filterIgnoreGroupsChanged, 8))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QQmlListProperty<QAbstractCollisionShape>*>(_v) = _t->collisionShapes(); break;
        case 1: *reinterpret_cast<bool*>(_v) = _t->sendContactReports(); break;
        case 2: *reinterpret_cast<bool*>(_v) = _t->receiveContactReports(); break;
        case 3: *reinterpret_cast<bool*>(_v) = _t->sendTriggerReports(); break;
        case 4: *reinterpret_cast<bool*>(_v) = _t->receiveTriggerReports(); break;
        case 5: *reinterpret_cast<int*>(_v) = _t->filterGroup(); break;
        case 6: *reinterpret_cast<int*>(_v) = _t->filterIgnoreGroups(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 1: _t->setSendContactReports(*reinterpret_cast<bool*>(_v)); break;
        case 2: _t->setReceiveContactReports(*reinterpret_cast<bool*>(_v)); break;
        case 3: _t->setSendTriggerReports(*reinterpret_cast<bool*>(_v)); break;
        case 4: _t->setReceiveTriggerReports(*reinterpret_cast<bool*>(_v)); break;
        case 5: _t->setfilterGroup(*reinterpret_cast<int*>(_v)); break;
        case 6: _t->setFilterIgnoreGroups(*reinterpret_cast<int*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QAbstractPhysicsNode::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QAbstractPhysicsNode::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QAbstractPhysicsNodeE_t>.strings))
        return static_cast<void*>(this);
    return QQuick3DNode::qt_metacast(_clname);
}

int QAbstractPhysicsNode::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QQuick3DNode::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    return _id;
}

// SIGNAL 0
void QAbstractPhysicsNode::bodyContact(QAbstractPhysicsNode * _t1, const QVector<QVector3D> & _t2, const QVector<QVector3D> & _t3, const QVector<QVector3D> & _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2, _t3, _t4);
}

// SIGNAL 1
void QAbstractPhysicsNode::sendContactReportsChanged(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void QAbstractPhysicsNode::receiveContactReportsChanged(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void QAbstractPhysicsNode::sendTriggerReportsChanged(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void QAbstractPhysicsNode::receiveTriggerReportsChanged(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void QAbstractPhysicsNode::enteredTriggerBody(QAbstractPhysicsNode * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void QAbstractPhysicsNode::exitedTriggerBody(QAbstractPhysicsNode * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void QAbstractPhysicsNode::filterGroupChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void QAbstractPhysicsNode::filterIgnoreGroupsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}
QT_WARNING_POP
